from abc import ABC, abstractmethod
import os
from pathlib import Path

import numpy as np
from skimage.io import imread, imsave

import czifile
import tifffile

from model_server.base.process import make_rgb
from model_server.base.process import is_mask

class GenericImageDataAccessor(ABC):

    @abstractmethod
    def __init__(self):
        """
        Abstract base class that exposes an interfaces for image data, irrespective of whether it is instantiated
        from file I/O or other means.  Enforces X, Y, C, Z dimensions in that order.
        """
        pass

    @property
    def chroma(self):
        return self.shape_dict['C']

    @staticmethod
    def conform_data(data):
        if len(data.shape) > 4 or (0 in data.shape):
            raise DataShapeError(f'Cannot handle image with dimensions other than X, Y, C, and Z: {data.shape}')
        ones = [1 for i in range(0, 4 - len(data.shape))]
        return data.reshape(*data.shape, *ones)

    def is_3d(self):
        return True if self.shape_dict['Z'] > 1 else False

    def is_mask(self):
        return is_mask(self._data)

    def get_one_channel_data (self, channel: int):
        c = int(channel)
        return InMemoryDataAccessor(self.data[:, :, c:(c+1), :])

    @property
    def pixel_scale_in_micrometers(self):
        return {}

    @property
    def dtype(self):
        return self.data.dtype

    @property
    def hw(self):
        """
        Get data height and width as a tuple
        :return: tuple of (Y, X) dimensions
        """
        return self.shape_dict['Y'], self.shape_dict['X']

    @property
    def nz(self):
        return self.shape_dict['Z']

    @property
    def data(self):
        """
        Return data as 4d with axes in order of Y, X, C, Z
        :return: np.ndarray
        """
        return self._data

    @property
    def shape(self):
        return self._data.shape

    @property
    def shape_dict(self):
        return dict(zip(('Y', 'X', 'C', 'Z'), self.data.shape))

class InMemoryDataAccessor(GenericImageDataAccessor):
    def __init__(self, data):
        self._data = self.conform_data(data)

class GenericImageFileAccessor(GenericImageDataAccessor): # image data is loaded from a file
    def __init__(self, fpath: Path):
        """
        Interface for image data that originates in an image file
        :param fpath: absolute path to image file
        :param kwargs: variable-length keyword arguments
        """
        if not os.path.exists(fpath):
            raise FileAccessorError(f'Could not find file at {fpath}')
        self.fpath = fpath

class TifSingleSeriesFileAccessor(GenericImageFileAccessor):
    def __init__(self, fpath: Path):
        super().__init__(fpath)

        try:
            tf = tifffile.TiffFile(fpath)
            self.tf = tf
        except Exception:
            FileAccessorError(f'Unable to access data in {fpath}')

        if len(tf.series) != 1:
            raise DataShapeError(f'Expect only one series in {fpath}')

        se = tf.series[0]

        order = ['Y', 'X', 'C', 'Z']
        axs = [a for a in se.axes if a in order]
        da = se.asarray()

        if 'C' not in axs:
            axs.append('C')
            da = np.expand_dims(da, len(da.shape))

        if 'Z' not in axs:
            axs.append('Z')
            da = np.expand_dims(da, len(da.shape))

        yxcz = np.moveaxis(
            da,
            [axs.index(k) for k in order],
            [0, 1, 2, 3]
        )

        self._data = self.conform_data(yxcz.reshape(yxcz.shape[0:4]))

    def __del__(self):
        self.tf.close()

class PngFileAccessor(GenericImageFileAccessor):
    def __init__(self, fpath: Path):
        super().__init__(fpath)

        try:
            arr = imread(fpath)
        except Exception:
            FileAccessorError(f'Unable to access data in {fpath}')

        if len(arr.shape) == 3: # rgb
            self._data = np.expand_dims(arr, 3)
        else: # mono
            self._data = np.expand_dims(arr, (2, 3))

class CziImageFileAccessor(GenericImageFileAccessor):
    """
    Image that is stored in a Zeiss .CZI file; may be multi-channel, and/or a z-stack,
    but not a time series or multiposition acquisition.
    """
    def __init__(self, fpath: Path):
        super().__init__(fpath)

        try:
            cf = czifile.CziFile(fpath)
            self.czifile = cf
        except Exception:
            raise FileAccessorError(f'Unable to access CZI data in {fpath}')

        sd = {ch: cf.shape[cf.axes.index(ch)] for ch in cf.axes}
        if (sd.get('S') and (sd['S'] > 1)) or (sd.get('T') and (sd['T'] > 1)):
            raise DataShapeError(f'Cannot handle image with multiple positions or time points: {sd}')

        idx = {k: sd[k] for k in ['Y', 'X', 'C', 'Z']}
        yxcz = np.moveaxis(
            cf.asarray(),
            [cf.axes.index(ch) for ch in idx],
            [0, 1, 2, 3]
        )
        self._data = self.conform_data(yxcz.reshape(yxcz.shape[0:4]))

    def __del__(self):
        self.czifile.close()

    @property
    def pixel_scale_in_micrometers(self):
        scale_meta = self.czifile.metadata(raw=False)['ImageDocument']['Metadata']['Scaling']['Items']['Distance']
        sc = {}
        for m in scale_meta:
            if m['DefaultUnitFormat'].encode() == b'\xc2\xb5m' and m['Id'] in self.shape_dict.keys():  # literal mu-m
                sc[m['Id']] = m['Value'] * 1e6
        return sc


def write_accessor_data_to_file(fpath: Path, acc: GenericImageDataAccessor, mkdir=True) -> bool:
    """
    Export an image accessor to file.
    :param fpath: complete path including filename and extension
    :param acc: image accessor to be written
    :param mkdir: create any needed subdirectories in fpath if True
    :return: True
    """
    if 'P' in acc.shape_dict.keys():
        raise FileWriteError(f'Can only write single-position accessor to file')
    ext = fpath.suffix.upper()

    if mkdir:
        fpath.parent.mkdir(parents=True, exist_ok=True)

    if ext == '.PNG':
        if acc.dtype != 'uint8':
            raise FileWriteError(f'Invalid data type {acc.dtype}')
        if acc.chroma == 1:
            data = acc.data[:, :, 0, 0]
        elif acc.chroma == 2:  # add a blank blue channel
            data = make_rgb(acc.data)
        else:  # preserve RGB order
            data = acc.data[:, :, :, 0]
        imsave(fpath, data, check_contrast=False)
        return True

    elif ext in ['.TIF', '.TIFF']:
        zcyx= np.moveaxis(
            acc.data,  # yxcz
            [3, 2, 0, 1],
            [0, 1, 2, 3]
        )
        if acc.is_mask():
            if acc.dtype == 'bool':
                data = (zcyx * 255).astype('uint8')
            else:
                data = zcyx.astype('uint8')
            tifffile.imwrite(fpath, data, imagej=True)
        else:
            tifffile.imwrite(fpath, zcyx, imagej=True)
    else:
        raise FileWriteError(f'Unable to write data to file of extension {ext}')
    return True


def generate_file_accessor(fpath):
    """
    Given an image file path, return an image accessor, assuming the file is a supported format and represents
    a single position array, which may be single or multi-channel, single plane or z-stack.
    """
    if str(fpath).upper().endswith('.TIF') or str(fpath).upper().endswith('.TIFF'):
        return TifSingleSeriesFileAccessor(fpath)
    elif str(fpath).upper().endswith('.CZI'):
        return CziImageFileAccessor(fpath)
    elif str(fpath).upper().endswith('.PNG'):
        return PngFileAccessor(fpath)
    else:
        raise FileAccessorError(f'Could not match a file accessor with {fpath}')


class PatchStack(InMemoryDataAccessor):

    def __init__(self, data):
        """
        A sequence of n (generally) color 3D images of the same size
        :param data: either a list of np.ndarrays of size YXCZ, or np.ndarray of size PYXCZ
        """

        if isinstance(data, list):  # list of YXCZ patches
            n = len(data)
            yxcz_shape = np.array([e.shape for e in data]).max(axis=0)
            nda = np.zeros(
                (n, *yxcz_shape), dtype=data[0].dtype
            )
            for i in range(0, len(data)):
                s = tuple([slice(0, c) for c in data[i].shape])
                nda[i][s] = data[i]

        elif isinstance(data, np.ndarray) and len(data.shape) == 5:  # interpret as PYXCZ
            nda = data
        else:
            raise InvalidDataForPatchStackError(f'Cannot create accessor from {type(data)}')

        assert nda.ndim == 5
        self._data = nda

    def iat(self, i):
        return InMemoryDataAccessor(self.data[i, :, :, :, :])

    def iat_yxcz(self, i):
        return self.iat(i)

    @property
    def count(self):
        return self.shape_dict['P']

    @property
    def shape_dict(self):
        return dict(zip(('P', 'Y', 'X', 'C', 'Z'), self.data.shape))

    def get_list(self):
        n = self.nz
        return [self.data[:, :, 0, zi] for zi in range(0, n)]

    @property
    def pyxcz(self):
        return self.data

    @property
    def pczyx(self):
        return np.moveaxis(
            self.data,
            [0, 3, 4, 1, 2],
            [0, 1, 2, 3, 4]
        )

    @property
    def shape(self):
        return self.data.shape

    @property
    def shape_dict(self):
        return dict(zip(('P', 'Y', 'X', 'C', 'Z'), self.data.shape))


def make_patch_stack_from_file(fpath):  # interpret z-dimension as patch position
    if not Path(fpath).exists():
        raise FileNotFoundError(f'Could not find {fpath}')

    pyxc = np.moveaxis(
        generate_file_accessor(fpath).data,  # yxcz
        [0, 1, 2, 3],
        [1, 2, 3, 0]
    )
    pyxcz = np.expand_dims(pyxc, axis=3)
    return PatchStack(pyxcz)



class Error(Exception):
    pass

class FileAccessorError(Error):
    pass

class FileNotFoundError(Error):
    pass

class DataShapeError(Error):
    pass

class FileWriteError(Error):
    pass

class InvalidAxisKey(Error):
    pass

class InvalidDataShape(Error):
    pass

class InvalidDataForPatchStackError(Error):
    pass



