from abc import ABC, abstractmethod
from math import floor

import numpy as np

from model_server.base.accessors import GenericImageDataAccessor, InMemoryDataAccessor


class Model(ABC):

    def __init__(self, autoload=True, params=None):
        """
        Abstract base class for an inference model that uses image data as an input.

        :param autoload: automatically load model and dependencies into memory if True
        :param params: Dict[str, str] of arguments e.g. configuration files required to load model
        """
        self.autoload = autoload
        self.params = params
        self.loaded = False
        if not autoload:
            return None
        if self.load():
            self.loaded = True
        else:
            raise CouldNotLoadModelError()
        return None

    @abstractmethod
    def load(self):
        """
        Abstract method that carries out the expectedly time-consuming step of loading a model into memory
        :return: True if successful, else False
        """
        pass

    @abstractmethod
    def infer(self, *args) -> (object, dict):
        """
        Abstract method that carries out the computationally intensive step of running data through a model
        :param args:
        :return:
        """
        pass

    def reload(self):
        self.load()


class ImageToImageModel(Model):
    """
    Abstract class for models that receive an image and return an image of the same size
    """

    @abstractmethod
    def infer(self, img: GenericImageDataAccessor) -> (GenericImageDataAccessor, dict):
        pass


class SemanticSegmentationModel(ImageToImageModel):
    """
    Base model that exposes a method that returns a binary mask for a given input image and pixel class
    """

    @abstractmethod
    def label_pixel_class(
            self, img: GenericImageDataAccessor, **kwargs) -> GenericImageDataAccessor:
        """
        Given an image, return an image of the same shape where each pixel is assigned to one or more integer classes
        """
        pass


class InstanceSegmentationModel(ImageToImageModel):
    """
    Base model that exposes a method that returns an instance classification map for a given input image and mask
    """

    @abstractmethod
    def label_instance_class(
            self, img: GenericImageDataAccessor, mask: GenericImageDataAccessor, **kwargs
    ) -> GenericImageDataAccessor:
        """
        Given an image and a mask of the same size, return a map where each connected object is assigned a class
        """
        if not mask.is_mask():
            raise InvalidInputImageError('Expecting a binary mask')
        if not img.shape == mask.shape:
            raise InvalidInputImageError('Expect input image and mask to be the same shape')


class DummySemanticSegmentationModel(SemanticSegmentationModel):

    model_id = 'dummy_make_white_square'

    def load(self):
        return True

    def infer(self, img: GenericImageDataAccessor) -> (GenericImageDataAccessor, dict):
        super().infer(img)
        w = img.shape_dict['X']
        h = img.shape_dict['Y']
        result = np.zeros([h, w], dtype='uint8')
        result[floor(0.25 * h) : floor(0.75 * h), floor(0.25 * w) : floor(0.75 * w)] = 255
        return InMemoryDataAccessor(data=result), {'success': True}

    def label_pixel_class(
            self, img: GenericImageDataAccessor, **kwargs) -> GenericImageDataAccessor:
        mask, _ = self.infer(img)
        return mask

class DummyInstanceSegmentationModel(InstanceSegmentationModel):

    model_id = 'dummy_pass_input_mask'

    def load(self):
        return True

    def infer(
            self, img: GenericImageDataAccessor, mask: GenericImageDataAccessor
    ) -> (GenericImageDataAccessor, dict):
        return img.__class__(
            (mask.data / mask.data.max()).astype('uint16')
        )

    def label_instance_class(
            self, img: GenericImageDataAccessor, mask: GenericImageDataAccessor, **kwargs
    ) -> GenericImageDataAccessor:
        """
        Returns a trivial segmentation, i.e. the input mask with value 1
        """
        super(DummyInstanceSegmentationModel, self).label_instance_class(img, mask, **kwargs)
        return self.infer(img, mask)

class Error(Exception):
    pass

class CouldNotLoadModelError(Error):
    pass

class ParameterExpectedError(Error):
    pass

class InvalidInputImageError(Error):
    pass